SystemOrganization addCategory: #'Neo-Caching'!

Object subclass: #NeoCacheStatistics
	instanceVariableNames: 'hits misses'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Caching'!

!NeoCacheStatistics commentStamp: '<historical>' prior: 0!
I am NeoCacheStatistics.
I keep track of hits and misses of a cache.!

!NeoCacheStatistics methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:01'!
addHit
	hits := hits + 1! !

!NeoCacheStatistics methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:00'!
addMiss
	misses := misses + 1! !

!NeoCacheStatistics methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:01'!
hitRatio
	"Return the ratio of hits against total calls I received.
	This will be a number between 0 and 1.
	When I am empty, return 0."
	
	| total |
	total := hits + misses.
	total = 0 ifTrue: [ ^ 0 ].
	^ hits / total! !

!NeoCacheStatistics methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 15:59'!
hits
	^ hits! !

!NeoCacheStatistics methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 15:59'!
initialize
	super initialize.
	hits := misses := 0! !

!NeoCacheStatistics methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 15:59'!
misses
	^ misses! !

!NeoCacheStatistics methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 16:00'!
reset
	hits := misses := 0! !

Object subclass: #NeoCacheWeight
	instanceVariableNames: 'total maximum compute'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Caching'!

!NeoCacheWeight commentStamp: '<historical>' prior: 0!
I  am NeoCacheWeight.
I keep track of the weight of a cache.

The weight of a cache is the sum of the weight of all values currently present. The simplest and default weight calculation returns a constant 1 for each value, effectively counting the number of values.

The default maximum is 16.

Using compute, a selector or block, applied to a value, different calculation can be made. Consider for example #sizeInMemory.!

!NeoCacheWeight methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:17'!
add: value
	| weight |
	weight := compute cull: value.
	total := total + weight! !

!NeoCacheWeight methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:25'!
compute
	^ compute! !

!NeoCacheWeight methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 16:18'!
compute: valuable
	compute := valuable! !

!NeoCacheWeight methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 16:16'!
initialize
	super initialize.
	total := 0.
	maximum := 16.
	compute := [ 1 ]! !

!NeoCacheWeight methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:27'!
isBelowMaximum
	^ total <= maximum ! !

!NeoCacheWeight methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:25'!
maximum
	^ maximum! !

!NeoCacheWeight methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 16:18'!
maximum: integer
	maximum := integer! !

!NeoCacheWeight methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:17'!
remove: value
	| weight |
	weight := compute cull: value.
	total := total - weight! !

!NeoCacheWeight methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 16:19'!
reset
	total := 0! !

!NeoCacheWeight methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:19'!
total
	^ total! !

Object subclass: #NeoDoubleLink
	instanceVariableNames: 'value nextLink previousLink'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Caching'!

!NeoDoubleLink commentStamp: '<historical>' prior: 0!
I am NeoDoubleLink, the elementary part of a double linked list.

I hold a value, as well as a link to my successor (nextLink) and to my predecessor (previousLink) - both can be nil.!

!NeoDoubleLink class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 12/2/2013 16:41'!
value: anObject
	^ self new
		value: anObject;
		yourself! !

!NeoDoubleLink methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/2/2013 20:57'!
asNeoDoubleLink
	^ self! !

!NeoDoubleLink methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:43'!
hasPredecessor
	^ previousLink notNil! !

!NeoDoubleLink methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:43'!
hasSuccessor
	^ nextLink notNil! !

!NeoDoubleLink methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:40'!
nextLink
	^ nextLink! !

!NeoDoubleLink methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:40'!
nextLink: anObject
	nextLink := anObject! !

!NeoDoubleLink methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:40'!
previousLink
	^ previousLink! !

!NeoDoubleLink methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:40'!
previousLink: anObject
	previousLink := anObject! !

!NeoDoubleLink methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 16:42'!
reset
	nextLink := previousLink := nil! !

!NeoDoubleLink methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:41'!
value
	^ value! !

!NeoDoubleLink methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:40'!
value: anObject
	value := anObject! !

Object subclass: #NeoDoubleLinkedList
	instanceVariableNames: 'head tail'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Caching'!

!NeoDoubleLinkedList commentStamp: '<historical>' prior: 0!
I am NeoDoubleLinkedList, an ordered list data structure consisting of objects, most likely NeoDoubleLinks or something compatible, connected to each other by forward and backwards links.!

!NeoDoubleLinkedList methodsFor: 'adding' stamp: 'SvenVanCaekenberghe 12/2/2013 21:08'!
add: anObjectOrLink
	^ self addLast: anObjectOrLink! !

!NeoDoubleLinkedList methodsFor: 'adding' stamp: 'SvenVanCaekenberghe 12/2/2013 21:16'!
add: anObjectOrLink afterLink: otherLink
	| link otherLinkSuccessor |
	otherLink = tail
		ifTrue: [ ^ self addLast: anObjectOrLink ].
	link := anObjectOrLink asNeoDoubleLink.
	otherLinkSuccessor := otherLink nextLink.
	otherLink nextLink: link.
	link previousLink: otherLink.
	link nextLink: otherLinkSuccessor.
	otherLinkSuccessor previousLink: link.
	^ link! !

!NeoDoubleLinkedList methodsFor: 'adding' stamp: 'SvenVanCaekenberghe 12/2/2013 21:20'!
add: anObjectOrLink beforeLink: otherLink
	| link otherLinkPredeccessor |
	otherLink = head
		ifTrue: [ ^ self addFirst: anObjectOrLink ].
	link := anObjectOrLink asNeoDoubleLink.
	otherLinkPredeccessor := otherLink previousLink.
	otherLink previousLink: link.
	link nextLink: otherLink.
	link previousLink: otherLinkPredeccessor.
	otherLinkPredeccessor nextLink: link.
	^ link! !

!NeoDoubleLinkedList methodsFor: 'adding' stamp: 'SvenVanCaekenberghe 12/2/2013 21:06'!
addFirst: anObjectOrLink
	| link |
	link := anObjectOrLink asNeoDoubleLink.
	link nextLink: head.
	head ifNotNil: [ head previousLink: link ].
	tail ifNil: [ tail := link ].
	head := link.
	^ link! !

!NeoDoubleLinkedList methodsFor: 'adding' stamp: 'SvenVanCaekenberghe 12/2/2013 21:07'!
addLast: anObjectOrLink
	| link |
	link := anObjectOrLink asNeoDoubleLink.
	link previousLink: tail.
	tail ifNotNil: [ tail nextLink: link ].
	head ifNil: [ head := link ].
	tail := link.
	^ link! !

!NeoDoubleLinkedList methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 12/2/2013 23:40'!
asArray
	^ Array streamContents: [ :out |
		self do: [ :each | out nextPut: each ] ]! !

!NeoDoubleLinkedList methodsFor: 'enumerating' stamp: 'SvenVanCaekenberghe 12/2/2013 19:09'!
do: block
	| current |
	current := head.
	[ current isNil ] 
		whileFalse: [ 
	 		block value: current value.
			current := current nextLink ]! !

!NeoDoubleLinkedList methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/3/2013 11:15'!
first
	self isEmpty ifTrue: [ CollectionIsEmpty signalWith: self ].
	^ head value! !

!NeoDoubleLinkedList methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 19:08'!
isEmpty
	^ head isNil and: [ tail isNil ]! !

!NeoDoubleLinkedList methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/3/2013 11:15'!
last
	self isEmpty ifTrue: [ CollectionIsEmpty signalWith: self ].
	^ tail value! !

!NeoDoubleLinkedList methodsFor: 'removing' stamp: 'SvenVanCaekenberghe 12/3/2013 20:10'!
removeFirst
	| link |
	self isEmpty ifTrue: [ CollectionIsEmpty signalWith: self ].
	link := head.
	head := head nextLink.
	head 
		ifNil: [ tail := nil ]
		ifNotNil: [ head previousLink: nil ].
	link reset.
	^ link! !

!NeoDoubleLinkedList methodsFor: 'removing' stamp: 'SvenVanCaekenberghe 12/3/2013 20:10'!
removeLast
	| link |
	self isEmpty ifTrue: [ CollectionIsEmpty signalWith: self ].
	link := tail.
	tail := tail previousLink.
	tail 
		ifNil: [ head := nil ]
		ifNotNil: [ tail nextLink: nil ].
	link reset.
	^ link! !

!NeoDoubleLinkedList methodsFor: 'removing' stamp: 'SvenVanCaekenberghe 12/2/2013 20:56'!
removeLink: link
	| predecessor successor |
	predecessor := link previousLink.
	successor := link nextLink.
	predecessor
		ifNil: [ head := successor ]
		ifNotNil: [ predecessor nextLink: successor ].
	successor
		ifNil: [ tail := predecessor ]
		ifNotNil: [ successor previousLink: predecessor ].
	link reset! !

Object subclass: #NeoLRUCache
	instanceVariableNames: 'lruHead lruTail keyIndex weight statistics factory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Caching'!

!NeoLRUCache commentStamp: '<historical>' prior: 0!
I am NeoLRUCache.

I am a limited cache that evicts the least recently used entries. My implementation is properly O(1).

My primary interface is #at:ifAbsentPut: which takes two arguments: a key and a block. Either the key is found (cache hit) and its value is returned, or the key is not found (cache miss). If the latter case, block should compute a new value to cache. Because block takes the key as optional argument, you can specify a factory style argument as well. With an explicit factory specified, you can also use #at: to access me.

For each addition to the cache, a weight is computed by #computeWeight (a selector or block) and added to #totalWeight. When totalWeight is no longer below maximumWeight, the least recently used item of the cache is evicted (removed) to make room. 

The default #computeWeight returns 1 for each value, effectively counting the number of entries. The default #maximumWeight is 16.

I count hits and misses and can return my #hitRatio.

Implementation Notes

The key/value pairs in the cache are held as Associations in a linked list, using ValueLink's, between lruHead and lruTail, ordered from least to most recently used.

The keyIndex Dictionary maps from each key to the predecessor ValueLink of the  ValueLink holding the matching key/value pair. To refer to the first element, lruHead, nil is used as predecessor.

New pairs are added at the end of the list.

In case of a hit, a pair gets promoted to the end of the list.

In case of a full cache, the first pair of the list gets evicted.

When manipulating the linked list, the predecessor pointers in keyIndex must be updated where necessary.

See #validateInvariantWith: where the relationship between the 2 datastructures is checked.!

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/2/2013 16:23'!
addWeightOf: value 
	weight add: value.
	[ weight isBelowMaximum ]
		whileFalse: [ self evict ]! !

!NeoLRUCache methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:02'!
at: key
	"If key is present in the cache, return the associated value.
	This is a hit and makes that key/value pair the most recently used.
	If key is absent, use the factory to compute a new value and cache it.
	This is a miss and will create a new key/value pair entry.
	Furthermore this could result in the least recently used key/value pair 
	being removed when the specified maximum cache weight is exceeded.
	A factory is required for this behavior."
	
	^ self 
		at: key 
		ifAbsentPut: (factory ifNil: [ self error: 'No factory specified' ])! !

!NeoLRUCache methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/30/2013 19:21'!
at: key ifAbsentPut: block
	"If key is present in the cache, return the associated value.
	This is a hit and makes that key/value pair the most recently used.
	If key is absent, use block to compute a new value and cache it.
	Block can optionally take one argument, the key. 
	This is a miss and will create a new key/value pair entry.
	Furthermore this could result in the least recently used key/value pair 
	being removed when the specified maximum cache weight is exceeded."

	| association |
	association := keyIndex 
		associationAt: key 
		ifAbsent: [ 
			| value |
			value := block cull: key.
			^ self handleMiss: key -> value ].
	^ self handleHit: association
		! !

!NeoLRUCache methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 16:26'!
computeWeight: valuable
	"Set the way to compute the weight of each cached value.
	This can be either a Symbol or one argument block.
	When the total weight is no longer below the maximum weight,
	the least recently used key/value pair will be removed.
	The default way to compute the weight returns 1 for each value,
	effectively counting the number of cached values."
	
	weight compute: valuable! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/30/2013 19:27'!
evict
	| link value |
	link := self removeFirstLink.
	value := link value.
	self removeWeightOf: value value.
	keyIndex removeKey: value key.
	lruHead 
		ifNotNil: [ keyIndex at: lruHead value key put: nil ]! !

!NeoLRUCache methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 11/29/2013 20:55'!
factory: block
	"Set the factory to compute values from keys to block.
	The factory will be evaluated for each key not present.
	Only my #at: message will use the factory."
	
	factory := block! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/2/2013 16:05'!
handleHit: association
	| predecessor newPredecessor link |
	statistics addHit.
	predecessor := association value.
	link := predecessor 
		ifNil: [ lruHead ] 
		ifNotNil: [ predecessor nextLink ].
	newPredecessor := self promoteSuccessorOf: predecessor.
	association value: newPredecessor.
	^ link value value! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/2/2013 16:05'!
handleMiss: association
	| link predecessor |
	statistics addMiss.
	link := association asLink.
	self addWeightOf: association value.
	predecessor := self insertLinkAtEnd: link.
	keyIndex at: association key put: predecessor.
	^ association value! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 12/2/2013 16:04'!
hitRatio
	"Return the ratio of hits against total calls I received.
	This will be a number between 0 and 1.
	When I am empty, return 0."
	
	^ statistics hitRatio! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 12/2/2013 16:04'!
hits
	"Return how many hits, requests for keys present I received."

	^ statistics hits! !

!NeoLRUCache methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:03'!
includesKey: key
	"Return true when the receiver has a value cached for key."
	
	^ keyIndex includesKey: key! !

!NeoLRUCache methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 16:24'!
initialize
	super initialize.
	keyIndex := Dictionary new.
	weight := NeoCacheWeight new.
	statistics := NeoCacheStatistics new! !

!NeoLRUCache methodsFor: 'private - linked list' stamp: 'SvenVanCaekenberghe 11/30/2013 19:55'!
insertLinkAtEnd: link
	| predecessor |
	lruTail 
		ifNil: [ 
			lruHead := lruTail := link.
			predecessor := nil ]
		ifNotNil: [
		 	predecessor := lruTail.
			lruTail nextLink: link.
			lruTail := link ].
	link nextLink: nil.
	^ predecessor! !

!NeoLRUCache methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:22'!
isEmpty
	"Return true when the receiver contains no entries."
	
	^ keyIndex isEmpty! !

!NeoLRUCache methodsFor: 'enumerating' stamp: 'SvenVanCaekenberghe 11/29/2013 21:19'!
keysAndValuesDo: block
	"Execute block with each key and value present in me.
	This will be from least to most recently used."
	
	| link |
	link := lruHead.
	[ link notNil ]
		whileTrue: [ 
			block 
				value: link value key 
				value: link value value.
			link := link nextLink ]! !

!NeoLRUCache methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/2/2013 16:26'!
maximumWeight: limit
	"Set my maximum allowed total weight of all cached values to limit.
	If the total weight is no longer below limit,
	the least recently used key/value pair will be removed.
	The default maximum weight limit is 16."
	
	weight maximum: limit! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 12/2/2013 16:05'!
misses
	"Return how many misses, requests for keys not present I received."
	
	^ statistics misses! !

!NeoLRUCache methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:24'!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	stream nextPut: $#; print: self size.
	stream space; print: weight total; nextPut: $/; print: weight maximum.
	stream space; print: weight compute.
	stream space; print: (self hitRatio * 100.0) rounded ; nextPut: $%.
	stream nextPut: $)! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/1/2013 00:13'!
promoteSuccessorOf: predecessor
	| link  |
	lruHead = lruTail ifTrue: [ ^ nil ].
	predecessor
		ifNil: [ 
			link := lruHead.
			lruHead := link nextLink ]
		ifNotNil: [ 
			link := predecessor nextLink.
			link = lruTail ifTrue: [ ^ predecessor ].
			predecessor nextLink: link nextLink ].
	keyIndex at: link nextLink value key put: predecessor.
	^ self insertLinkAtEnd: link! !

!NeoLRUCache methodsFor: 'removing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:25'!
removeAll
	"Remove all key/value pairs that I currently hold,
	effectiley resetting me, but not my statistics."
	
	keyIndex removeAll.
	weight reset.
	lruHead := lruTail := nil! !

!NeoLRUCache methodsFor: 'private - linked list' stamp: 'SvenVanCaekenberghe 11/30/2013 19:26'!
removeFirstLink
	^ self removeLink: nil! !

!NeoLRUCache methodsFor: 'removing' stamp: 'SvenVanCaekenberghe 11/29/2013 21:23'!
removeKey: key
	"If I currently cache key, remove the entry.
	Signal a KeyNotFound when I currently do not cache key.
	Return the removed value."
	
	^ self 
		removeKey: key 
		ifAbsent: [ KeyNotFound signalFor: key in: self ]! !

!NeoLRUCache methodsFor: 'removing' stamp: 'SvenVanCaekenberghe 11/30/2013 19:33'!
removeKey: key ifAbsent: block
	"If I currently cache key, remove the entry.
	Execute block when key is currently absent.
	Return the removed value."
	
	^ (self includesKey: key)
		ifTrue: [ 
			| predecessor link |
			predecessor := keyIndex at: key.
			link := self removeLink: predecessor.
			(predecessor notNil and: [ predecessor nextLink notNil ])
				ifTrue: [ keyIndex at: predecessor nextLink value key put: predecessor ].
			self removeWeightOf: link value value.
			keyIndex removeKey: key.
			link value value ]
		ifFalse: block! !

!NeoLRUCache methodsFor: 'private - linked list' stamp: 'SvenVanCaekenberghe 11/30/2013 19:31'!
removeLastLink
	| link |
	link := lruHead.
	lruHead := lruTail := nil.
	^ link! !

!NeoLRUCache methodsFor: 'private - linked list' stamp: 'SvenVanCaekenberghe 11/30/2013 19:31'!
removeLink: predecessor
	| link |
	lruHead = lruTail
		ifTrue: [ 
			^ self removeLastLink ].
	predecessor 
		ifNil: [ 
			link := lruHead.
			lruHead := link nextLink ]
		ifNotNil: [
			link := predecessor nextLink.
			predecessor nextLink: link nextLink ].
	link nextLink: nil.
	^ link! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/2/2013 16:25'!
removeWeightOf: value 
	weight remove: value! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 11/29/2013 21:15'!
size
	"Return the count of items currently present."
	
	^ keyIndex size! !

!NeoLRUCache methodsFor: 'accessing - statistics' stamp: 'SvenVanCaekenberghe 12/2/2013 16:25'!
totalWeight
	"Return the total weight of all cached values currently present."
	
	^ weight total! !

!NeoLRUCache methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/28/2013 23:01'!
validateInvariantWith: assertable
	keyIndex keysAndValuesDo: [ :key :predecessor |
		predecessor 
			ifNil: [ 
				assertable assert: lruHead value key = key ]
			ifNotNil: [ 
				assertable assert: predecessor nextLink value key = key ] ].
	self keysAndValuesDo: [ :key :value |
		(keyIndex at: key)
			ifNil: [ 
				assertable assert: lruHead value value = value.
				assertable assert: lruHead value key = key ]
			ifNotNil: [ :predecessor | 
				assertable assert: predecessor nextLink value value = value.
				assertable assert: predecessor nextLink value key = key ] ]! !

!Object methodsFor: '*Neo-Caching' stamp: 'SvenVanCaekenberghe 12/2/2013 20:58'!
asNeoDoubleLink
	^ NeoDoubleLink value: self! !

TestCase subclass: #NeoDoubleLinkedListTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Caching'!

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 11:40'!
testAddAfterLink
	| list link |
	list := NeoDoubleLinkedList new.
	list add: 1.
	link := list add: 2.
	list add: 3.
	list add: 0 afterLink: link.
	self assert: list asArray equals: #( 1 2 0 3 )! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 11:40'!
testAddBeforeLink
	| list link |
	list := NeoDoubleLinkedList new.
	list add: 1.
	link := list add: 2.
	list add: 3.
	list add: 0 beforeLink: link.
	self assert: list asArray equals: #( 1 0 2 3 )! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 11:31'!
testAddFirst
	| list |
	list := NeoDoubleLinkedList new.
	1 to: 3 do: [ :each | list add: each ].
	list addFirst: 0.
	self deny: list isEmpty.
	self assert: list first equals: 0.
	self assert: list last equals: 3.
	self assert: list asArray equals: #( 0 1 2 3 )! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 11:31'!
testAddLast
	| list |
	list := NeoDoubleLinkedList new.
	1 to: 3 do: [ :each | list add: each ].
	list addLast: 0.
	self deny: list isEmpty.
	self assert: list first equals: 1.
	self assert: list last equals: 0.
	self assert: list asArray equals: #( 1 2 3 0 )! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 11:18'!
testEmpty
	| list |
	list := NeoDoubleLinkedList new.
	self assert: list isEmpty.
	self should: [ list first ] raise: CollectionIsEmpty.
	self should: [ list last ] raise: CollectionIsEmpty.
	list do: [ :each | self fail ]
	! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 11:24'!
testOne
	| list |
	list := NeoDoubleLinkedList new.
	list add: #one.
	self deny: list isEmpty.
	self assert: list first equals: #one.
	self assert: list last equals: #one.
	self assert: list asArray equals: #( one )! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 20:11'!
testOneRemoveFirst
	| list |
	list := NeoDoubleLinkedList new.
	list add: #one.
	list removeFirst.
	self assert: list isEmpty.
	list add: #one.
	list removeLast.
	self assert: list isEmpty! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 20:11'!
testOneRemoveLast
	| list |
	list := NeoDoubleLinkedList new.
	list add: #one.
	list removeLast.
	self assert: list isEmpty.
	list add: #one.
	list removeLast.
	self assert: list isEmpty! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 11:26'!
testThree
	| list |
	list := NeoDoubleLinkedList new.
	1 to: 3 do: [ :each | list add: each ].
	self deny: list isEmpty.
	self assert: list first equals: 1.
	self assert: list last equals: 3.
	self assert: list asArray equals: #( 1 2 3 )! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 20:06'!
testThreeRemoveFirst
	| list link |
	list := NeoDoubleLinkedList new.
	1 to: 3 do: [ :each | list add: each ].
	link := list removeFirst.
	self assert: link value equals: 1.
	self assert: list first equals: 2.
	self assert: list last equals: 3.
	self assert: list asArray equals: #( 2 3 )! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 20:07'!
testThreeRemoveLast
	| list link |
	list := NeoDoubleLinkedList new.
	1 to: 3 do: [ :each | list add: each ].
	link := list removeLast.
	self assert: link value equals: 3.
	self assert: list first equals: 1.
	self assert: list last equals: 2.
	self assert: list asArray equals: #( 1 2 )! !

!NeoDoubleLinkedListTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/3/2013 20:12'!
testThreeRemoveLink
	| list link |
	list := NeoDoubleLinkedList new.
	list add: 1.
	link := list add: 2.
	list add: 3.
	list removeLink: link.
	self assert: list first equals: 1.
	self assert: list last equals: 3.
	self assert: list asArray equals: #( 1 3 )! !

TestCase subclass: #NeoLRUCacheTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Caching'!

!NeoLRUCacheTests commentStamp: '<historical>' prior: 0!
I am NeoLRUCacheTests.
!

!NeoLRUCacheTests methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:31'!
newCache
	^ NeoLRUCache new! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:31'!
test5kClasses
	| cache data |
	(cache := self newCache)
		maximumWeight: 1024.
	data := Object allSubclasses first: 5*1024.
	data do: [ :each |
		cache at: each name ifAbsentPut: [ each ] ].
	cache validateInvariantWith: self.
	self assert: cache size equals: 1024.
	data shuffled do: [ :each |
		cache at: each name ifAbsentPut: [ each ] ].
	self assert: cache size equals: 1024.
	cache validateInvariantWith: self.	
	data 
		select: [ :each | cache includesKey: each name ]
		thenDo: [ :each | cache at: each name ifAbsentPut: [ self fail ] ].
	self assert: cache hits >= 1024.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:32'!
test6k
	| cache |
	(cache := self newCache)
		maximumWeight: 600.
	1 to: 6000 do: [ :each |
		cache at: each ifAbsentPut: [ each * 2 ] ].
	self assert: cache size equals: 600.
	self assert: (cache includesKey: 5401).
	self deny: (cache includesKey: 5400).
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:32'!
testCustomWeight
	| cache |
	(cache := self newCache)
		computeWeight: #sizeInMemory;
		maximumWeight: 64.
	cache at: 1 ifAbsentPut: [ ByteArray new: 1 ].
	self assert: cache totalWeight equals: (ByteArray new: 1) sizeInMemory.
	2 to: 10 do: [ :each | 
		cache at: each ifAbsentPut: [ ByteArray new: each ] ].
	self assert: cache size equals: 3.
	self 
		assert: cache totalWeight 
		equals: ((8 to: 10) collect: [ :each | (ByteArray new: each) sizeInMemory ]) sum.
	self assert: (cache at: 10 ifAbsentPut: [ self fail ]) equals: (ByteArray new: 10).
	self deny: (cache includesKey: 1)! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:32'!
testEmpty
	| emptyCache |
	emptyCache := self newCache.
	self assert: emptyCache isEmpty.
	self assert: emptyCache size isZero.
	self assert: emptyCache hits isZero.
	self assert: emptyCache misses isZero.
	self assert: emptyCache hitRatio isZero.
	self assert: emptyCache totalWeight isZero.
	self deny: (emptyCache includesKey: #foo).
	emptyCache keysAndValuesDo: [ :key :value | self fail ]! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:32'!
testEnumeration
	| cache data keys values |
	cache := self newCache.
	data := Dictionary new.
	1 to: 10 do: [ :each | data at: each asWords put: each ].
	data keysAndValuesDo: [ :key :value |
		cache at: key ifAbsentPut: [ value ] ].
	keys := Array new writeStream.
	values := Array new writeStream.
	cache keysAndValuesDo: [ :key :value |
		self assert: (data at: key) equals: value.
		keys nextPut: key.
		values nextPut: value ].
	self assert: keys contents asSet equals: data keys asSet.
	self assert: values contents asSet equals: data values asSet.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:32'!
testEviction
	| cache |
	cache := self newCache.
	cache maximumWeight: 16.
	1 to: 20 do: [ :each |
		cache at: each asWords ifAbsentPut: [ each ] ].
	self assert: cache size equals: 16.
	self assert: cache totalWeight equals: 16.
	5 to: 20 do: [ :each |
		self assert: (cache includesKey: each asWords).
		self 
			assert: (cache at: each asWords ifAbsentPut: [ self fail ]) 
			equals: each ].
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:32'!
testFactory
	| cache |
	cache := self newCache.
	cache maximumWeight: 5.
	cache factory: [ :key | key * 2 ].
	#( 1 2 3 4 1 5 6 7 8 1 ) do: [ :each |
		 cache at: each ].
	self assert: cache size equals: 5.
	self assert: cache hits equals: 2.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:32'!
testFactoryStyle
	| cache factory |
	cache := self newCache.
	cache maximumWeight: 5.
	factory := [ :key | key * 2 ].
	#( 1 2 3 4 1 5 6 7 8 1 ) do: [ :each |
		 cache at: each ifAbsentPut: factory ].
	self assert: cache size equals: 5.
	self assert: cache hits equals: 2.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:32'!
testFixedAccess
	| cache data keys |
	(cache := self newCache)
		maximumWeight: 4.
	data := (1 to: 16) collect: #asWords.
	data do: [ :each |
		cache at: each ifAbsentPut: [ each ] ].
	cache validateInvariantWith: self.
	self assert: cache size equals: 4.
	(#(14 15) collect: #asWords) do: [ :each |
		cache at: each ifAbsentPut: [ each ] ].
	self assert: cache size equals: 4.
	cache validateInvariantWith: self.	
	keys := data select: [ :each | cache includesKey: each ].
	keys do: [ :each | cache at: each ifAbsentPut: [ self fail ] ].
	self assert: cache hits >= 4.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:32'!
testOne
	| cache |
	cache := self newCache.
	cache at: #foo ifAbsentPut: [ 100 ].
	self assert: (cache includesKey: #foo).
	self deny: cache isEmpty.
	self assert: cache size equals: 1.
	self assert: cache totalWeight equals: 1.
	self assert: cache hits isZero.
	self assert: cache misses equals: 1.
	self assert: cache hitRatio isZero.
	self deny: (cache includesKey: #bar).
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:33'!
testOneHit
	| cache value |
	cache := self newCache.
	cache at: #foo ifAbsentPut: [ 100 ].
	self assert: (cache includesKey: #foo).
	self assert: cache hits equals: 0.
	self assert: cache misses equals: 1.
	value := cache at: #foo ifAbsentPut: [ self fail ].
	self assert: value equals: 100.
	self assert: cache hits equals: 1.
	self assert: cache misses equals: 1.
	cache validateInvariantWith: self
	! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:33'!
testOneHitTwice
	| cache value |
	cache := self newCache.
	cache at: #foo ifAbsentPut: [ 100 ].
	value := cache at: #foo ifAbsentPut: [ self fail ].
	self assert: value equals: 100.
	value := cache at: #foo ifAbsentPut: [ self fail ].
	self assert: value equals: 100.
	self assert: cache hitRatio equals: 2/3.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:33'!
testPrimeFactors
	"[ self run: #testPrimeFactors ] bench."
	
	| cache data |
	cache := self newCache.
	cache maximumWeight: 512.
	cache factory: [ :key | key ].
	data := Array streamContents: [ :out |
		1 to: 4096 do: [ :each |
			each primeFactorsOn: out.
			out nextPut: each ] ].
	data := data collect: [ :each | each asWords ].
	data do: [ :each | cache at: each ].
	self assert: cache totalWeight equals: 512.
	self assert: cache hitRatio > (7/10).
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:33'!
testRandomAccess
	| cache data random keys |
	(cache := self newCache)
		maximumWeight: 4.
	data := (1 to: 16) collect: #asWords.
	data do: [ :each |
		cache at: each ifAbsentPut: [ each ] ].
	cache validateInvariantWith: self.
	self assert: cache size equals: 4.
	random := data shuffled.
	random do: [ :each |
		cache at: each ifAbsentPut: [ each ] ].
	self assert: cache size equals: 4.
	cache validateInvariantWith: self.	
	keys := data select: [ :each | cache includesKey: each ].
	keys do: [ :each | cache at: each ifAbsentPut: [ self fail ] ].
	self assert: cache hits >= 4.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:33'!
testRemoveAll
	| cache |
	cache := self newCache.
	1 to: 10 do: [ :each | 
		cache at: each ifAbsentPut: [ each ] ].
	cache removeAll.
	self assert: cache isEmpty.
	self assert: cache size isZero.
	self assert: cache totalWeight isZero.
	self assert: cache misses equals: 10.
	self assert: cache hits isZero.
	self deny: (cache includesKey: #1).
	cache keysAndValuesDo: [ :key :value | self fail ].
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:33'!
testRemoveOne
	| cache |
	cache := self newCache.
	cache at: #foo ifAbsentPut: [ 100 ].
	cache removeKey: #foo.
	self assert: cache isEmpty.
	self assert: cache size isZero.
	self assert: cache totalWeight isZero.
	self deny: (cache includesKey: #foo).
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:33'!
testRemoveOneOfThree
	| cache |
	cache := self newCache.
	cache at: #x ifAbsentPut: [ 100 ].
	cache at: #y ifAbsentPut: [ 200 ].
	cache at: #z ifAbsentPut: [ 300 ].
	cache removeKey: #y.
	self assert: cache size equals: 2.
	self assert: cache totalWeight equals: 2.
	self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100.
	self deny: (cache includesKey: #y).
	self assert: (cache at: #z ifAbsentPut: [ self fail ]) equals: 300.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:33'!
testTen
	| cache |
	cache := self newCache.
	1 to: 10 do: [ :each |
		cache at: each asWords ifAbsentPut: [ each ] ].
	self assert: cache size equals: 10.
	1 to: 10 do: [ :each |
		self assert: (cache includesKey: each asWords) ].
	1 to: 10 do: [ :each |
		self 
			assert: (cache at: each asWords ifAbsentPut: [ self fail ]) 
			equals: each ].
	self assert: cache hitRatio equals: 1/2.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:34'!
testThree
	| cache |
	cache := self newCache.
	cache at: #x ifAbsentPut: [ 100 ].
	cache at: #y ifAbsentPut: [ 200 ].
	cache at: #z ifAbsentPut: [ 300 ].
	self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100.
	self assert: (cache at: #y ifAbsentPut: [ self fail ]) equals: 200.
	self assert: (cache at: #z ifAbsentPut: [ self fail ]) equals: 300.
	self assert: cache hitRatio equals: 3/6.
	self assert: (cache at: #z ifAbsentPut: [ self fail ]) equals: 300.
	self assert: (cache at: #y ifAbsentPut: [ self fail ]) equals: 200.
	self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100.
	self assert: cache hitRatio equals: 6/9.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:34'!
testThreeHitSameOne
	| cache |
	cache := self newCache.
	cache at: #x ifAbsentPut: [ 100 ].
	cache at: #y ifAbsentPut: [ 200 ].
	cache at: #z ifAbsentPut: [ 300 ].
	self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100.
	self assert: (cache at: #y ifAbsentPut: [ self fail ]) equals: 200.
	self assert: (cache at: #z ifAbsentPut: [ self fail ]) equals: 300.
	3 timesRepeat: [ 	
		self assert: (cache at: #x ifAbsentPut: [ self fail ]) equals: 100 ].
	self assert: cache hits equals: 6.
	cache validateInvariantWith: self! !

!NeoLRUCacheTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/2/2013 16:34'!
testTwo
	| cache |
	cache := self newCache.
	cache at: #foo ifAbsentPut: [ 100 ].
	cache at: #bar ifAbsentPut: [ 200 ].
	self assert: (cache at: #foo ifAbsentPut: [ self fail ]) equals: 100.
	self assert: (cache at: #bar ifAbsentPut: [ self fail ]) equals: 200.
	self assert: cache hitRatio equals: 2/4.
	self assert: (cache at: #bar ifAbsentPut: [ self fail ]) equals: 200.
	self assert: (cache at: #foo ifAbsentPut: [ self fail ]) equals: 100.
	self assert: cache hitRatio equals: 4/6.
	cache validateInvariantWith: self! !
